from enum import Enum
from typing import NamedTuple

from .device_api import DeviceApiABC, DeviceId


class Version(NamedTuple):
    major: int
    minor: int
    revision: int


class SystemInfoCommand(Enum):
    get_main_application_version = 0x00
    get_bootloader_version = 0x01
    get_main_application_build_info = 0x02
    get_board_revision = 0x03
    config_block_write = 0x04
    config_block_write_complete = 0x05
    get_mac_address = 0x06
    get_user_configblock = 0x07
    set_user_configblock = 0x08
    set_user_configblock_complete = 0x09
    disable_factory_mode = 0x0a
    get_log_status = 0x0b
    get_log_chunk = 0x0c
    clear_log = 0x0d
    get_nordic_temperature = 0x0e
    startup_async = 0x0f
    get_model_number = 0x12
    get_stats_id = 0x13
    set_device_mode = 0x14
    get_device_mode = 0x15
    get_device_mode_async = 0x16
    get_secondary_mcu_version = 0x17
    get_secondary_mcu_version_async = 0x18
    get_animation_version = 0x19
    get_animation_version_async = 0x1a
    set_audio_crc = 0x22
    get_audio_crc = 0x23
    get_level_one_diagnostics = 0x26
    level_one_diagnostics_async = 0x27
    get_sku_LMQ = 0x28 # value from Core library used by LMQ
    get_secondary_mcu_status = 0x29
    secondary_mcu_status_async = 0x2a
    get_sku = 0x38


class SystemInfo(DeviceApiABC):
    device_id = DeviceId.system_info

    def get_main_application_version(self) -> Version:
        """
        Get version of toys firmware

        :return Version:
        """

        response = self.request(SystemInfoCommand.get_main_application_version)
        return Version(
            major=int.from_bytes(response.data[:2], "big"),
            minor=int.from_bytes(response.data[2:4], "big"),
            revision=int.from_bytes(response.data[4:], "big"),
        )

    def get_bootloader_version(self) -> Version:
        """
        Get version of toys bootloader

        :return Version:
        """

        response = self.request(SystemInfoCommand.get_bootloader_version)
        return Version(
            major=int.from_bytes(response.data[:2], "big"),
            minor=int.from_bytes(response.data[2:4], "big"),
            revision=int.from_bytes(response.data[4:], "big"),
        )

    def get_mac_address(self) -> str:
        """
        Get toy's mac address as "aa:bb:cc:dd:ee:ff"

        :return str:
        """

        response = self.request(SystemInfoCommand.get_mac_address)
        return ":".join(chr(b1) + chr(b2) for b1, b2 in zip(response.data[0::2], response.data[1::2]))

    def get_nordic_temperature(self) -> int:
        """
        ???

        :return str: Celsius temperature
        """

        response = self.request(SystemInfoCommand.get_nordic_temperature)
        return int.from_bytes(response.data, "big") // 4

    def get_stats_id(self) -> int:
        """
        ???

        :return:
        """

        response = self.request(SystemInfoCommand.get_stats_id, target_id=0x11)
        return int.from_bytes(response.data, "big")

    def get_sku(self) -> str:
        response = self.request(SystemInfoCommand.get_sku, target_id=0x11)
        return "".join(chr(b) for b in response.data)



from enum import Enum

from pysphero.helpers import UnknownEnumMixing


class GenericCharacteristic(Enum):
    device_name = 0x2a00
    client_characteristic_configuration = 0x2902
    peripheral_preferred_connection_parameters = 0x2a04


class SpheroCharacteristic(Enum):
    force_band = "00020005-574f-4f20-5370-6865726f2121"
    api_v2 = "00010002-574f-4f20-5370-6865726f2121"


class Api2Error(UnknownEnumMixing, Enum):
    success = 0x00
    bad_device_id = 0x01
    bad_command_id = 0x02
    not_yet_implemented = 0x03
    command_is_restricted = 0x04
    bad_data_length = 0x05
    command_failed = 0x06
    bad_parameter_value = 0x07
    busy = 0x08
    bad_target_id = 0x09
    target_unavailable = 0x0a
    unknown = 0xff


class Toy(Enum):
    unknown = "Unknown"

    sphero_sprk = "Sphero SPRK+"
    sphero_bolt = "Sphero Bolt"
    sphero_mini = "Sphero MINI"
    ollie = "Ollie"
    r2d2 = "R2D2"
    r2q5 = "R2Q5"
    bb8 = "BB8"
    bb9e = "BB9E"
    lmq = "LightningMcQueen"


TOY_BY_PREFIX = {
    "SK-": Toy.sphero_sprk,
    "SB-": Toy.sphero_bolt,
    "SM-": Toy.sphero_mini,
    "2B-": Toy.ollie,
    "D2-": Toy.r2d2,
    "Q5-": Toy.r2q5,
    "BB-": Toy.bb8,
    "GB-": Toy.bb9e,
    "LM-": Toy.lmq,
}



import logging
from typing import NamedTuple, ClassVar

from pysphero.bluetooth import BleAdapter
from pysphero.bluetooth.ble_adapter import AbstractBleAdapter
from pysphero.constants import Toy
from pysphero.device_api import Animatronics, Sensor, UserIO, ApiProcessor, Power, SystemInfo
from pysphero.driving import Driving
from pysphero.exceptions import PySpheroException
from pysphero.helpers import cached_property

logger = logging.getLogger(__name__)


class PeripheralPreferredConnectionParameters(NamedTuple):
    min_con_interval: int
    max_con_interval: int
    slave_latency: int
    connection_supervision_timeout_multiplier: int


class Sphero:
    """
    High-level API for communicate with sphero toy
    """

    def __init__(
            self,
            mac_address: str,
            toy_type: Toy = Toy.unknown,
            ble_adapter_cls: ClassVar[AbstractBleAdapter] = BleAdapter
    ):
        self.mac_address = mac_address
        self.type = toy_type
        self._ble_adapter_cls = ble_adapter_cls
        self._ble_adapter = None

    @property
    def ble_adapter(self):
        if self._ble_adapter is None:
            raise PySpheroException("Use Sphero as context manager")
        return self._ble_adapter

    @ble_adapter.setter
    def ble_adapter(self, value):
        self._ble_adapter = value

    def __enter__(self):
        self._ble_adapter = self._ble_adapter_cls(self.mac_address)
        # if self.type is Toy.unknown:
        #     self.type = TOY_BY_PREFIX.get(self.name[:3], Toy.unknown)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ble_adapter.close()

    @cached_property
    def system_info(self) -> SystemInfo:
        return SystemInfo(ble_adapter=self.ble_adapter)

    @cached_property
    def power(self) -> Power:
        return Power(ble_adapter=self.ble_adapter)

    @cached_property
    def driving(self) -> Driving:
        return Driving(ble_adapter=self.ble_adapter)

    @cached_property
    def api_processor(self) -> ApiProcessor:
        return ApiProcessor(ble_adapter=self.ble_adapter)

    @cached_property
    def user_io(self) -> UserIO:
        return UserIO(ble_adapter=self.ble_adapter)

    @cached_property
    def sensor(self) -> Sensor:
        return Sensor(ble_adapter=self.ble_adapter)

    @cached_property
    def animatronics(self) -> Animatronics:
        return Animatronics(ble_adapter=self.ble_adapter)



import logging
from typing import Optional

import gatt

from pysphero.bluetooth.ble_adapter import AbstractBleAdapter
from pysphero.bluetooth.packet_collector import PacketCollector
from pysphero.constants import SpheroCharacteristic
from pysphero.exceptions import PySpheroRuntimeError
from pysphero.packet import Packet

logger = logging.getLogger(__name__)


class Device(gatt.Device):

    def __init__(self, packet_collector: PacketCollector, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.packet_collector = packet_collector

    def characteristic_value_updated(self, characteristic, value):
        """
        Callback from gatt when value was updated, call the create packet function for received data.
        """
        self.packet_collector.append_raw_data(value)


class GattAdapter(AbstractBleAdapter):
    def __init__(self, mac_address):
        logger.debug("Init Gatt Adapter")
        super().__init__(mac_address)

        self.manager = gatt.DeviceManager("hci0")

        self._device = Device(
            self.packet_collector, mac_address=self.mac_address, manager=self.manager)
        self._device.connect()

        ch_force_band = self._find_characteristic(SpheroCharacteristic.force_band.value)
        ch_force_band.write_value(b"usetheforce...band")

        self.ch_api_v2 = self._find_characteristic(SpheroCharacteristic.api_v2.value)

        self.ch_api_v2.enable_notifications()
        self._executor.submit(self.manager.run)

    def _find_characteristic(self, characteristic: str):
        found_characteristic = None
        self._device.services_resolved()
        for service in self._device.services:
            for ch in service.characteristics:
                if str(ch.uuid) == characteristic:
                    found_characteristic = ch

        if not found_characteristic:
            raise PySpheroRuntimeError(f"Characteristic {characteristic} not found")

        return found_characteristic

    def close(self):
        self.manager.stop()
        self._device.disconnect()
        super().close()

    def write(self, packet: Packet, *, timeout: float = 10, raise_api_error: bool = True) -> Optional[Packet]:
        """
         Method allow send request packet and get response packet

         :param packet: request packet
         :param timeout: timeout waiting for a response from sphero
         :param raise_api_error: raise exception when receive api error
         :return Packet: response packet
        """
        logger.debug(f"Send {packet}")
        self.ch_api_v2.write_value(packet.build())
        return self.packet_collector.get_response(packet, raise_api_error, timeout)



from enum import Enum


class SystemModeCommand(Enum):
    set_system_mode = 0x00
    get_system_mode = 0x01
    enable_system_mode_notification = 0x02
    system_mode_changed = 0x03
    set_max_speed = 0x04
    get_max_speed = 0x05
    set_current_weapon = 0x06
    get_current_weapon = 0x07
    set_aiming_mode = 0x08
    get_playmode_unlock_mask = 0x09
    set_playmode_unlock_mask = 0x0a
    enable_menu_item_change_notification = 0x0b
    menu_item_changed = 0x0c
    get_enabled_weapons_mask = 0x0d
    set_enabled_weapons_mask = 0x0e
    get_holocron_counts = 0x10
    dec_holocron_count = 0x13
    clear_holocron_count = 0x14
    find_holocron_now = 0x15
    enable_force_awareness_reporting = 0x16
    force_awareness_event_async = 0x17
    set_category_rarity = 0x18
    get_angle_to_holocron = 0x19
    set_fa_disturbance_timeout = 0x1a
    set_audio_header = 0x1b
    set_audio_header_async = 0x1c
    enable_robot_bas_app_report = 0x20
    robot_bad_app_report = 0x21
    get_pending_action_list = 0x22
    get_audio_info = 0x23



from pysphero.constants import Api2Error


class PySpheroException(Exception):
    ...


class PySpheroRuntimeError(PySpheroException):
    ...


class PySpheroTimeoutError(PySpheroException):
    ...


class PySpheroApiError(PySpheroException):
    def __init__(self, api_error: Api2Error):
        super().__init__(f"Api reponse error: {api_error.name}")


class PySpheroNotFoundError(PySpheroException):
    ...



import struct
from itertools import zip_longest
from typing import Callable


class cached_property:
    def __init__(self, func: Callable):
        self.func = func

    def __get__(self, instance, type=None):
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result


class UnknownEnumMixing:
    unknown = 0x00

    @classmethod
    def _missing_(cls, value):
        return cls.unknown


def float_from_bytes(data):
    return struct.unpack(">f", bytearray(data))[0]


def grouper(iterable, n, fillvalue=None):
    """Collect data into fixed-length chunks or blocks"""
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)



import abc
import logging
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from typing import Callable, Optional

from pysphero.bluetooth.packet_collector import PacketCollector
from pysphero.exceptions import PySpheroRuntimeError
from pysphero.packet import Packet

logger = logging.getLogger(__name__)

STOP_NOTIFY = object()


class AbstractBleAdapter(abc.ABC):
    def __init__(self, mac_address, max_workers=2):
        self.mac_address = mac_address
        self.packet_collector = PacketCollector()

        self._executor = ThreadPoolExecutor(max_workers=max_workers)
        self._running = Event()  # disable receiver thread
        self._running.set()
        self._notify_future = None

    def close(self):
        self._running.clear()
        self._executor.shutdown(wait=False)

    def write(self, packet: Packet, *, timeout: float = 10, raise_api_error: bool = True) -> Optional[Packet]:
        """
         Method allow send request packet and get response packet

         :param packet: request packet
         :param timeout: timeout waiting for a response from sphero
         :param raise_api_error: raise exception when receive api error
         :return Packet: response packet
         """

    def start_notify(self, packet: Packet, callback: Callable, timeout: float = 10):
        self._notify_future = self._executor.submit(
            self.notify_worker,
            callback,
            packet,
            timeout,
        )
        return self._notify_future

    def stop_notify(self):
        if self._notify_future is None:
            raise PySpheroRuntimeError("Future not found")

        logger.debug(f"[NOTIFY_WORKER] Cancel")
        self._notify_future.cancel()
        self._notify_future = None

    def notify_worker(self, callback: Callable, packet: Packet, timeout: float):
        logger.debug(f"[NOTIFY_WORKER] Start {packet}")

        while self._running.is_set():
            response = self.packet_collector.get_response(packet, raise_api_error=False, timeout=timeout)
            logger.debug(f"[NOTIFY_WORKER] Received {response}")
            if callback(response) is STOP_NOTIFY:
                logger.debug(f"[NOTIFY_WORKER] Received STOP_NOTIFY")
                self._notify_future = None
                break



from enum import Enum


class WifiCommand(Enum):
    power_on_wifi = 0x00
    power_on_wifi_complete = 0x01
    get_wifi_mac = 0x02
    start_update = 0x03
    erase_block_range = 0x04
    erase_block_range_complete = 0x05
    send_data = 0x06
    send_data_complete = 0x07
    update_done = 0x08
    update_done_complete = 0x09



import contextlib
import logging
from typing import List, Optional

from bluepy.btle import DefaultDelegate, Peripheral, ADDR_TYPE_RANDOM, Characteristic, Descriptor

from pysphero.bluetooth.ble_adapter import AbstractBleAdapter
from pysphero.bluetooth.packet_collector import PacketCollector
from pysphero.constants import SpheroCharacteristic, GenericCharacteristic
from pysphero.packet import Packet

logger = logging.getLogger(__name__)


class BluepyDelegate(DefaultDelegate):
    """
    Delegate class for bluepy
    Getting bytes from peripheral and build packets
    """

    def __init__(self, packet_collector: PacketCollector):
        super().__init__()
        self.packet_collector = packet_collector

    def handleNotification(self, handle: int, data: List[int]):
        """
        handleNotification getting raw data from peripheral and save it.
        This function may be called several times. Therefore, the state is stored inside the class.

        :param handle:
        :param data: raw data
        :return:
        """
        self.packet_collector.append_raw_data(data)


class BluepyAdapter(AbstractBleAdapter):
    STOP_NOTIFY = object()

    def __init__(self, mac_address):
        logger.debug("Init Bluepy Adapter")
        super().__init__(mac_address)
        self.delegate = BluepyDelegate(self.packet_collector)
        self.peripheral = Peripheral(self.mac_address, ADDR_TYPE_RANDOM)
        self.peripheral.setDelegate(self.delegate)

        self.ch_api_v2 = self._get_characteristic(uuid=SpheroCharacteristic.api_v2.value)
        # Initial api descriptor
        # Need for getting response from sphero
        desc = self._get_descriptor(self.ch_api_v2, GenericCharacteristic.client_characteristic_configuration.value)
        desc.write(b"\x01\x00", withResponse=True)

        self._executor.submit(self._receiver)
        logger.debug("Bluepy Adapter: successful initialization")

    def close(self):
        super().close()
        # ignoring any exception
        # because it does not matter
        with contextlib.suppress(Exception):
            self.peripheral.disconnect()

    def write(self, packet: Packet, *, timeout: float = 10, raise_api_error: bool = True) -> Optional[Packet]:
        logger.debug(f"Send {packet}")
        self.ch_api_v2.write(packet.build(), withResponse=True)
        return self.packet_collector.get_response(packet, raise_api_error, timeout=timeout)

    def _receiver(self):
        logger.debug("Start receiver")

        sleep = 0.05
        while self._running.is_set():
            self.peripheral.waitForNotifications(sleep)

        logger.debug("Stop receiver")

    def _get_characteristic(self, uuid: int or str) -> Characteristic:
        return self.peripheral.getCharacteristics(uuid=uuid)[0]

    def _get_descriptor(self, characteristic: Characteristic, uuid: int or str) -> Descriptor:
        return characteristic.getDescriptors(forUUID=uuid)[0]



from enum import Enum
from typing import NamedTuple, Callable, Type, Tuple, List

from pysphero.helpers import float_from_bytes, grouper
from pysphero.packet import Packet

from .device_api import DeviceApiABC, DeviceId


class SensorParameter(NamedTuple):
    flag: int
    min_value: float
    max_value: float
    modifier: Callable = None


class _Sensor(Enum):
    """Parent class for all sensors"""

    @classmethod
    def mask(cls) -> int:
        _mask = 0x0
        for parameter in cls:
            _mask |= parameter.value.flag

        return _mask


# next sensors class created for Sphero Bolt
# todo: refactoring this for all Sphero Toys
class Quaternion(_Sensor):
    x = SensorParameter(0x2000000, -1.0, 1.0)
    y = SensorParameter(0x1000000, -1.0, 1.0)
    z = SensorParameter(0x800000, -1.0, 1.0)
    w = SensorParameter(0x400000, -1.0, 1.0)


class Attitude(_Sensor):
    pitch = SensorParameter(0x40000, -179.0, 180.0)
    roll = SensorParameter(0x20000, -179.0, 180.0)
    yaw = SensorParameter(0x10000, -179.0, 180.0)


class Accelerometer (_Sensor):
    x = SensorParameter(0x8000, -8.19, 8.19)
    y = SensorParameter(0x4000, -8.19, 8.19)
    z = SensorParameter(0x2000, -8.19, 8.19)


class AccelOne(_Sensor):
    accel_one = SensorParameter(0x200, 0.0, 8000.0)


class Locator(_Sensor):
    x = SensorParameter(0x40, -32768.0, 32767.0, lambda x: x * 100)
    y = SensorParameter(0x20, -32768.0, 32767.0, lambda x: x * 100)


class Velocity(_Sensor):
    x = SensorParameter(0x10, -32768.0, 32767.0, lambda x: x * 100)
    y = SensorParameter(0x08, -32768.0, 32767.0, lambda x: x * 100)


class Speed(_Sensor):
    speed = SensorParameter(0x04, 0.0, 32767.0)


class CoreTime(_Sensor):
    core_time = SensorParameter(0x02, 0.0, 0.0)


class Gyroscope(_Sensor):
    x = SensorParameter(0x2000000, -20000.0, 20000.0)
    y = SensorParameter(0x1000000, -20000.0, 20000.0)
    z = SensorParameter(0x800000, -20000.0, 20000.0)


class AmbientLight(_Sensor):
    ambient_light = SensorParameter(0x40000, 0.0, 120000.0)


class SensorCommand(Enum):
    set_sensor_streaming_mask = 0x00
    get_sensor_streaming_mask = 0x01
    sensor_streaming_data = 0x02
    run_accel_gyro_self_test = 0x06
    set_extended_sensor_streaming_mask = 0x0c
    get_extended_sensor_streaming_mask = 0x0d
    set_gyro_max_async = 0x0f
    gyro_max_async = 0x10
    configure_collision_detection = 0x11
    collision_detected_async = 0x12
    reset_locator = 0x13
    enable_collision_detected_async = 0x14
    subscribe_maneuver_async_notification = 0x15
    maneuver_detection_async = 0x16
    set_locator_flags = 0x17
    enable_accelerometer_activity_notify = 0x19
    accelerometer_activity_notify = 0x1a
    magnetometer_calibrate_to_north = 0x25
    magnetometer_north_yaw_notify = 0x26
    get_ambient_light_sensor_value = 0x30


class Sensor(DeviceApiABC):
    device_id = DeviceId.sensors

    def _set_sensor_streaming_mask(self, mask, interval: int = 250, count: int = 0):
        self.request(
            command_id=SensorCommand.set_sensor_streaming_mask,
            data=[*interval.to_bytes(2, "big"),  # interval
                  count & 0xff,  # count,
                  *mask.to_bytes(4, "big"),  # 0x00, 0x00, 0x80, 0x00  # mask
                  ],
            target_id=0x12,
        )

    def set_notify(
            self,
            callback: Callable,
            *sensors: Type[_Sensor],
            interval: int = 250,
            count: int = 0,
            timeout: float = 1,
    ):
        parameters = []
        mask = 0x0
        for sensor in sensors:
            parameters.extend(parameter for parameter in sensor)
            mask |= sensor.mask()

        parameters = sorted(parameters, key=lambda p: p.value.flag, reverse=True)

        def callback_wrapper(response: Packet):
            callback_data = {}
            for i, data in enumerate(grouper(response.data, 4, fillvalue=0x00)):
                value: float = float_from_bytes(data)

                parameter: Enum = parameters[i]
                if parameter.value.modifier:
                    value = parameter.value.modifier(value)

                callback_data[parameter] = value

            return callback(callback_data)

        self.notify(SensorCommand.sensor_streaming_data, callback_wrapper, timeout=timeout)
        self._set_sensor_streaming_mask(mask, interval, count)

    def cancel_notify_sensors(self):
        self.cancel_notify()

    def get_sensor_streaming_mask(self) -> Tuple[int, int, List[SensorParameter]]:
        response = self.request(
            command_id=SensorCommand.get_sensor_streaming_mask,
            target_id=0x12,
        )

        interval = int.from_bytes(response.data[:2], "big")
        count = int.from_bytes(response.data[2:3], "big")
        mask = int.from_bytes(response.data[3:], "big")

        parameters = []
        for sensor in _Sensor.__subclasses__():
            for parameter in sensor:
                if parameter.value.flag & mask:
                    parameters.append(parameter)

        return interval, count, parameters

    def get_ambient_light_sensor_value(self) -> float:
        response = self.request(
            command_id=SensorCommand.get_ambient_light_sensor_value,
            target_id=0x12,
        )

        return float_from_bytes(response.data)

    def magnetometer_calibrate_to_north(self):
        self.request(
            command_id=SensorCommand.magnetometer_calibrate_to_north,
            target_id=0x12,
        )



import abc
from concurrent.futures import Future
from enum import Enum
from typing import Callable

from pysphero.packet import Packet
from pysphero.packet import Flag


class DeviceId(Enum):
    api_processor = 0x10
    system_info = 0x11
    system_modes = 0x12
    power = 0x13
    driving = 0x16
    animatronics = 0x17
    sensors = 0x18
    peer_connection = 0x19
    user_io = 0x1a
    storage_command = 0x1b
    secondary_mcu_firmware_update_command = 0x1d
    wifi_command = 0x1e
    factory_test = 0x1f
    macro_system = 0x20
    proto = 0xfe


class DeviceApiABC(abc.ABC):
    device_id: Enum = NotImplemented

    def __init__(self, ble_adapter):
        self.ble_adapter = ble_adapter

    def request(self, command_id: Enum, timeout: float = 10, raise_api_error: bool = True, **kwargs) -> Packet:
        return self.ble_adapter.write(
            self.packet(command_id=command_id.value, **kwargs),
            raise_api_error=raise_api_error,
            timeout=timeout,
        )

    def notify(
            self,
            command_id: Enum,
            callback: Callable,
            timeout: float = 10,
            **kwargs
    ) -> Future:
        return self.ble_adapter.start_notify(
            self.packet(command_id=command_id.value, **kwargs),
            callback=callback,
            timeout=timeout,
        )

    def cancel_notify(self):
        self.ble_adapter.stop_notify()

    def packet(self, **kwargs):
        packet = Packet(
            device_id=self.device_id.value,
            **kwargs
        )
        return packet



from enum import Enum

from pysphero.device_api import DeviceApiABC, DeviceId
from pysphero.packet import Flag


class Direction(Enum):
    forward = 0x00
    reverse = 0x01


class TankDriveDirection(Enum):
    forward = 0x00
    in_place = 0x08
    reverse = 0x18


class DirectionRawMotor(Enum):
    disable = 0x00
    forward = 0x01
    reverse = 0x02


class StabilizationIndex(Enum):
    no_control_system = 0x00
    full_control_system = 0x01
    pitch_control_system = 0x02
    roll_control_system = 0x03
    yaw_control_system = 0x04
    speed_and_yaw_control_system = 0x05


class DrivingCommand(Enum):
    raw_motor = 0x01
    set_ackermann_steering_parameters = 0x02
    drift = 0x03
    absolute_yaw_steering = 0x04
    enable_flip_drive = 0x05
    reset_yaw = 0x06
    drive_with_heading = 0x07
    tank_drive = 0x08
    rc_car_drive = 0x09
    drive_to_position = 0x0a
    set_stabilization = 0x0c


class Driving(DeviceApiABC):
    device_id = DeviceId.driving

    def drive_with_heading(self, speed: int, heading: int, direction: Direction = Direction.forward):
        """

        :param int speed: speed from 0 to 255
        :param int heading: heading from 0 to 360
        :param Direction direction: motor rotation direction
        :return:
        """
        speed &= 0xff
        heading = heading.to_bytes(2, "big")
        direction = direction.value

        self.request(
            DrivingCommand.drive_with_heading,
            target_id=0x12,
            data=[speed, *heading, direction],
            flags=Flag.requests_response.value | Flag.command_has_target_id.value | Flag.resets_inactivity_timeout.value
        )

    def set_stabilization(self, stabilization_index: StabilizationIndex):
        """
        ???

        :param StabilizationIndex stabilization_index:
        :return:
        """
        self.request(
            DrivingCommand.set_stabilization,
            target_id=0x12,
            data=[stabilization_index.value],
        )

    def raw_motor(
            self,
            left_speed: int = 0x00,
            left_direction: DirectionRawMotor = DirectionRawMotor.forward,
            right_speed=0x00,
            right_direction: DirectionRawMotor = DirectionRawMotor.forward,
    ):
        """
        Control of each motor separately

        note: it works strange :) after wake up need sleeps

        :param int left_speed: speed of left motor from 0 to 255
        :param DirectionRawMotor left_direction:
        :param int right_speed: speed of right motor from 0 to 255
        :param DirectionRawMotor right_direction:
        :return:

        """
        left_direction = left_direction.value
        right_direction = right_direction.value

        self.request(
            DrivingCommand.raw_motor,
            target_id=0x12,
            data=[
                left_direction, left_speed & 0xff,
                right_direction, right_speed & 0xff,
            ],
        )

    def reset_yaw(self):
        """
        Reset of toys yaw

        :return:
        """
        self.request(
            DrivingCommand.reset_yaw,
            target_id=0x12,
        )

    def tank_drive(
            self,
            left_speed: int = 0x00,
            right_speed=0x00,
            direction: TankDriveDirection = TankDriveDirection.forward,
    ):
        """
        ??? todo: this method not exist in android app
        :param left_speed:
        :param right_speed:
        :param direction:
        :return:
        """
        self.request(
            DrivingCommand.tank_drive,
            data=[left_speed, right_speed, direction.value],
            target_id=0x12,
        )

    def ackermann_drive(
            self,
            steering: int = 0x00,
            direction: int = 0x00
    ):
        self.request(
            DrivingCommand.set_ackermann_steering_parameters,
            data=[*steering.to_bytes(4, "big"), *direction.to_bytes(4, "big")],
            flags=Flag.resets_inactivity_timeout.value
        )

    def ackermann_reset(
            self,
            steering: int = 0x00,
            direction: int = 0x00
    ):
        self.request(
            DrivingCommand.absolute_yaw_steering,
            data=[*steering.to_bytes(4, "big"), *direction.to_bytes(4, "big")],
            flags=Flag.resets_inactivity_timeout.value
        )



from enum import Enum

from .device_api import DeviceApiABC, DeviceId


class ApiProcessorCommand(Enum):
    echo = 0x00


class ApiProcessor(DeviceApiABC):
    device_id = DeviceId.api_processor

    def echo(self):
        """
        Send ping
        :return None:
        """

        self.request(ApiProcessorCommand.echo)



import logging
from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty
from threading import Event
from time import time
from typing import NamedTuple, Generator

from bluepy.btle import DefaultDelegate, Scanner, ScanEntry

from pysphero.constants import Toy, TOY_BY_PREFIX
from pysphero.core import Sphero
from pysphero.exceptions import PySpheroNotFoundError

logger = logging.getLogger(__name__)


class _ScanItem(NamedTuple):
    mac_address: str
    toy_type: Toy
    name: str


class _ScanDelegate(DefaultDelegate):

    def __init__(self):
        super().__init__()
        self.queue = Queue()

    def handleDiscovery(self, dev: ScanEntry, isNewDev: bool, isNewData: bool):
        name = dev.getValue(ScanEntry.COMPLETE_LOCAL_NAME) or ""
        toy_type = TOY_BY_PREFIX.get(name[:3])
        if toy_type:
            self.queue.put_nowait(_ScanItem(dev.addr, toy_type, name))


class _ContextScanner(Scanner):
    def __enter__(self, passive=False):
        self.clear()
        self.start(passive=passive)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


def _queue_iter(queue: Queue, timeout: float) -> Generator[_ScanItem, None, None]:
    stop_time = time() + timeout
    while time() <= stop_time:
        try:
            scan_item: _ScanItem = queue.get(timeout=0.2)
        except Empty:
            continue

        logger.debug("Found toy %s", scan_item)
        yield scan_item


def _scanner(delegate: _ScanDelegate, timeout: float, event: Event):
    stop_time = time() + timeout
    scanner = _ContextScanner().withDelegate(delegate)
    with scanner:
        while event.is_set() and time() <= stop_time:
            scanner.process(0.2)


def toy_scanner(*, toy_type: Toy = None, name: str = None, timeout: float = 5.0) -> Sphero:
    delegate = _ScanDelegate()
    running_event = Event()
    running_event.set()

    with ThreadPoolExecutor(max_workers=1) as executor:
        executor.submit(_scanner, delegate, timeout, running_event)
        for scan_item in _queue_iter(delegate.queue, timeout):

            # none when time is out
            if not scan_item:
                break

            if (
                    (not name and not toy_type) or
                    (name and name == scan_item.name) or
                    (toy_type and scan_item.toy_type == toy_type)
            ):
                running_event.clear()
                return Sphero(
                    mac_address=scan_item.mac_address,
                    toy_type=scan_item.toy_type,
                )

    raise PySpheroNotFoundError("Toy not found")



from .device_api import DeviceApiABC, DeviceId
from .animatronics import R2D2Animation, R2Q5Animation, BB9EAnimation, R2LegAction, LMQAnimation, Animatronics
from .api_processor import ApiProcessor
from .power import Power, BatteryVoltageStates, ChargerStates
from .sensor import Quaternion, Attitude, Accelerometer, AccelOne, \
    Locator, Velocity, Speed, CoreTime, Gyroscope, AmbientLight, Sensor
from .system_info import SystemInfo, Version
from .user_io import UserIO, Color, Pixel, Led, FrameRotation



from enum import Enum
from typing import NamedTuple

from .device_api import DeviceApiABC, DeviceId


class Color(NamedTuple):
    red: int = 0x00
    green: int = 0x00
    blue: int = 0x00

    def to_list(self):
        return [self.red & 0xff, self.green & 0xff, self.blue & 0xff]


class Pixel(NamedTuple):
    x: int = 0
    y: int = 0

    def to_list(self):
        return [self.x & 0x07, self.y & 0x07]


class Led(Enum):
    front_red = 0x01
    front_green = 0x02
    front_blue = 0x04
    back_red = 0x08
    back_green = 0x10
    back_blue = 0x20
    all = front_red | front_green | front_blue | back_red | back_green | back_blue


class FrameRotation(Enum):
    normal = 0x00
    degrees_90 = 0x01
    degrees_180 = 0x02
    degrees_270 = 0x03


class UserIOCommand(Enum):
    enable_gesture_event_notification = 0x00
    gesture_event = 0x01
    enable_button_event_notification = 0x02
    button_event = 0x03
    set_led = 0x04
    release_led_request = 0x05
    play_haptic_pattern = 0x06
    play_audio_file = 0x07
    set_audio_volume = 0x08
    get_audio_volume = 0x09
    stop_all_audio = 0x0a
    cap_touch_enable = 0x0b
    ambient_light_sensor_enable = 0x0c
    enable_ir = 0x0d
    set_all_leds = 0x0e
    set_backlight_intensity = 0x0f
    cap_touch_indication = 0x10
    enable_debug_data = 0x11
    assert_lcd_reset_pin = 0x12
    set_headlights = 0x13
    set_taillights = 0x14
    play_test_tone = 0x18
    start_idle_led = 0x19
    toy_commands = 0x20
    toy_events = 0x21
    set_user_profile = 0x22
    get_user_profile = 0x23
    set_all_leds_32_bit_mask = 0x1a
    set_all_leds_64_bit_mask = 0x1b
    set_all_leds_8_bit_mask = 0x1c
    set_led_matrix_pixel = 0x2d
    set_led_matrix_one_color = 0x2f
    set_led_matrix_frame_rotation = 0x3a
    set_led_matrix_text_scrolling = 0x3b
    set_led_matrix_text_scrolling_notify = 0x3c
    set_led_matrix_single_character = 0x42


class UserIO(DeviceApiABC):
    device_id = DeviceId.user_io

    def set_all_leds_8_bit_mask(
            self,
            front_color: Color = Color(),
            back_color: Color = Color(),
    ):
        """
        Set leds colors

        note: this command is supported Sphero Bolt.
        For another toys use set_all_leds

        :param Color front_color:
        :param Color back_color:
        :return:
        """
        leds = Led.all.value
        self.request(
            command_id=UserIOCommand.set_all_leds_8_bit_mask,
            data=[leds, *front_color.to_list(), *back_color.to_list()],
            target_id=0x11,
        )

    def set_led_matrix_one_color(self, color: Color = Color()):
        """
        Set matrix one color

        :param Color color:
        :return:
        """
        self.request(
            command_id=UserIOCommand.set_led_matrix_one_color,
            data=color.to_list(),
            target_id=0x12,
        )

    def set_led_matrix_pixel(self, pixel: Pixel, color: Color = Color()):
        """
        Put into pixel the color

        :param Pixel pixel:
        :param Color color:
        :return:
        """
        self.request(
            command_id=UserIOCommand.set_led_matrix_pixel,
            data=[*pixel.to_list(), *color.to_list()],
            target_id=0x12,
        )

    def set_led_matrix_single_character(self, symbol: str, color: Color):
        """
        Print symbol on matrix

        :param String symbol:
        :param Color color:
        :return:
        """
        self.request(
            command_id=UserIOCommand.set_led_matrix_single_character,
            data=[*color.to_list(), ord(symbol)],
            target_id=0x12,
        )

    def set_led_matrix_text_scrolling(self, string: str, color: Color, speed: int = 0x10, repeat: bool = True):
        """
        Print text on matrix

        :param str string: max length 6 symbols
        :param Color color:
        :param int speed: max value is 0x1e (30)
        :param bool repeat:
        :return:
        """
        self.request(
            command_id=UserIOCommand.set_led_matrix_text_scrolling,
            data=[
                *color.to_list(),
                speed % 0x1e,
                int(repeat),
                *[ord(c) for c in string[:7]],
                0x00,  # end line
            ],
            target_id=0x12,
        )

    def set_led_matrix_text_scrolling_notify(self):
        self.request(
            command_id=UserIOCommand.set_led_matrix_text_scrolling_notify,
            data=[0x00],  # todo: unknown
            target_id=0x11,
        )

    def set_led_matrix_frame_rotation(self, rotation: FrameRotation = FrameRotation.normal):
        self.request(
            command_id=UserIOCommand.set_led_matrix_frame_rotation,
            data=[rotation.value],
            target_id=0x11,
        )



from enum import Enum
from typing import List, Tuple

from pysphero.constants import Api2Error
from pysphero.exceptions import PySpheroRuntimeError
from pysphero.helpers import cached_property


class Flag(Enum):
    """
    Flag bits
    """
    response = 0x1
    requests_response = 0x2
    requests_only_error_response = 0x4
    resets_inactivity_timeout = 0x8
    command_has_target_id = 0x10
    command_has_source_id = 0x20


class Packet:
    """
    Packet structure:
    ---------------------------------
    - start      [1 byte]
    - flags      [1 byte]
    - source_id  [1 byte] (optional)
    - target_id  [1 byte] (optional)
    - device_id  [1 byte]
    - command_id [1 byte]
    - data       [n byte]
    - checksum   [1 byte]
    - end        [1 byte]
    ---------------------------------
    Usually the first data byte is the api_v2 response code
    """

    _sequence = 0x00

    start = 0x8d
    end = 0xd8

    escape = 0xab
    escape_mask = 0x88
    escaped_bytes = start & ~escape_mask, end & ~escape_mask, escape & ~escape_mask
    bad_bytes = start, end, escape

    def __init__(
            self,
            device_id: int,
            command_id: int,
            flags: int = None,
            target_id: int = None,
            source_id: int = None,
            sequence: int = None,
            data: List[int] = None,
    ):
        # For LMQ a notification with flag 0x00 is sent once audio has been played
        self.flags = flags
        if flags is None:
            self.flags = (Flag.requests_response.value | Flag.resets_inactivity_timeout.value)
        if flags is None and source_id is not None:
            self.flags |= Flag.command_has_source_id.value
        if flags is None and target_id is not None:
            self.flags |= Flag.command_has_target_id.value

        self.target_id = target_id
        self.source_id = source_id
        self.device_id = device_id
        self.command_id = command_id
        self.sequence = sequence if sequence is not None else self.generate_sequence()
        self.data = data or []
    
    @classmethod
    def generate_sequence(cls):
        """
        Autoincrement sequence number of packet
        """
        cls._sequence = (cls._sequence + 1) % 256
        return cls._sequence

    @property
    def id(self) -> Tuple:
        return self.device_id, self.command_id

    def __str__(self) -> str:
        return f"Packet(flg: {self.flags:#04x} did: {self.device_id:#04x} cid: {self.command_id:#04x} " \
               f"seq: {self.sequence:#04x})"

    def __repr__(self) -> str:
        return f"Packet(flg: {self.flags:#04x} tid: {self.target_id or 0x0:#04x} sid: {self.source_id or 0x0:#04x} " \
               f"did: {self.device_id:#04x} cid: {self.command_id:#04x} " \
               f"seq: {self.sequence:#04x}) data: {[hex(i) for i in self.data]} chs: {self.checksum:#04x})"

    @cached_property
    def api_error(self) -> Api2Error:
        if self.flags & Flag.response.value and len(self.data) > 0:
            return Api2Error(self.data.pop(0))  # delete first byte from data
        return Api2Error.success

    @staticmethod
    def _unescape_response_data(response_data) -> List[int]:
        raw_data = []

        iter_response_data = iter(response_data)
        for b in iter_response_data:
            # escaping byte allowing escaping start/end/escaping bytes
            if b == Packet.escape:
                # next byte is escaping
                b = next(iter_response_data, None)
                if b not in Packet.escaped_bytes:
                    raise PySpheroRuntimeError(f"Bad escaping byte {b:#04x}")

                b |= Packet.escape_mask

            raw_data.append(b)

        return raw_data

    @classmethod
    def from_response(cls, response_data: List[int]) -> "Packet":
        """
        Create packet from raw data
        :param response_data: raw data from peripheral
        :return Packet: response packet
        """
        response_data = Packet._unescape_response_data(response_data)
        start, flags, *data, checksum, end = response_data
        if start != cls.start or end != cls.end:
            raise PySpheroRuntimeError(
                f"Bad response packet: wrong start or end byte (start: {start:#04x}, end: {end:#04x})"
            )

        target_id = None
        if flags & Flag.command_has_target_id.value:
            target_id = data.pop(0)

        source_id = None
        if flags & Flag.command_has_source_id.value:
            source_id = data.pop(0)

        device_id, command_id, sequence, *data = data

        packet = cls(
            flags=flags,
            target_id=target_id,
            source_id=source_id,
            device_id=device_id,
            command_id=command_id,
            sequence=sequence,
            data=data,
        )

        calc_checksum = packet.checksum
        if calc_checksum != checksum:
            raise PySpheroRuntimeError(
                f"Bad response checksum. (Expected: {checksum:#04x}, obtained: {calc_checksum:#04x})"
            )

        return packet

    @property
    def packet_payload(self) -> List[int]:
        head = [self.flags]

        if self.target_id is not None:
            head.append(self.target_id)

        if self.source_id is not None:
            head.append(self.source_id)

        return [
            *head,
            self.device_id,
            self.command_id,
            self.sequence,
            *self.data,
        ]

    @property
    def checksum(self) -> int:
        return 0xff - (sum(self.packet_payload) & 0xff)

    def build(self) -> bytes:
        full_packet = [
            *self.packet_payload,
            self.checksum,
        ]

        escaped_full_packet = []
        for i in full_packet:
            if i in self.bad_bytes:
                escaped_full_packet.extend([self.escape, i & ~self.escape_mask])
            else:
                escaped_full_packet.append(i)

        return b"".join(i.to_bytes(1, byteorder="big") for i in [self.start, *escaped_full_packet, self.end])



from enum import Enum


class PeerConnectionCommand(Enum):
    enable_peer_connection_event_notification = 0x00
    peer_connection_event = 0x01
    get_peer_connection_state = 0x02
    set_bluetooth_name = 0x03
    get_bluetooth_name = 0x04



from enum import Enum

from pysphero.helpers import UnknownEnumMixing

from .device_api import DeviceApiABC, DeviceId


class BatteryVoltageStates(UnknownEnumMixing, Enum):
    ok = 0x01
    low = 0x02
    critical = 0x03


class ChargerStates(UnknownEnumMixing, Enum):
    not_charging = 0x01
    charging = 0x02
    charged = 0x03


class PowerCommand(Enum):
    enter_deep_sleep = 0x00
    enter_soft_sleep = 0x01
    get_usb_state = 0x02
    get_battery_voltage = 0x03
    get_battery_state_LMQ = 0x04  # value from Core library Used by LMQ
    enable_battery_state_change_notification = 0x05
    battery_state_changed_LMQ = 0x06  # value from Core library used by LMQ
    wake = 0x0d
    get_battery_percentage = 0x10
    set_power_options = 0x12
    get_power_options = 0x13
    get_battery_state = 0x17
    will_sleep_async = 0x19
    sleep_async = 0x1a
    battery_state_changed = 0x1f


class Power(DeviceApiABC):
    device_id = DeviceId.power

    def enter_deep_sleep(self):
        """
        Send shutdown command to toy

        :return None:
        """

        self.request(PowerCommand.enter_deep_sleep)

    def enter_soft_sleep(self):
        """
        Send sleep command to toy (normal condition of toy)

        :return None:
        """

        self.request(PowerCommand.enter_soft_sleep)

    def get_battery_voltage(self) -> float:
        """
        Returns battery voltage. Allows to determine the level of charge

        :float return: battery voltage in volts
        """

        response = self.request(PowerCommand.get_battery_voltage)
        return int.from_bytes(response.data, "big") / 100

    def wake(self):
        """
        Wake up toys

        :return None:
        """

        self.request(PowerCommand.wake)

    def get_battery_state(self) -> BatteryVoltageStates:
        """
        Get battery state without known voltage constants

        :return BatteryVoltageStates:
        """

        response = self.request(PowerCommand.get_battery_state)
        return BatteryVoltageStates(response.data[0])

    def battery_state_changed(self) -> ChargerStates:
        """
        Charging status information

        :return ChargerStates:
        """

        response = self.request(PowerCommand.battery_state_changed)
        return ChargerStates(response.data[0])



import struct
import time
from enum import Enum
from typing import Optional, List

from pysphero.helpers import float_from_bytes
from pysphero.packet import Packet

from .device_api import DeviceApiABC, DeviceId


class AnimatronicsCommand(Enum):
    play_animation = 0x05
    perform_leg_action = 0x0d
    set_head_position = 0x0f
    play_animation_complete_notify = 0x11
    get_head_position = 0x14
    set_leg_position = 0x15
    get_leg_position = 0x16
    get_leg_action = 0x25
    leg_action_complete_notify = 0x26
    stop_animation = 0x2b
    get_trophy_mode_enabled = 0x2e


class R2D2Animation(Enum):
    charger_1 = 0x00
    charger_2 = 0x01
    charger_3 = 0x02
    charger_4 = 0x03
    charger_5 = 0x04
    charger_6 = 0x05
    charger_7 = 0x06
    emote_alarm = 0x07
    emote_angry = 0x08
    emote_annoyed = 0x09
    emote_chatty = 0x0a
    emote_drive = 0x0b
    emote_excited = 0x0c
    emote_happy = 0x0d
    emote_ion_blast = 0x0e
    emote_laugh = 0x0f
    emote_no = 0x10
    emote_sad = 0x11
    emote_sassy = 0x12
    emote_scared = 0x13
    emote_spin = 0x14
    emote_yes = 0x15
    emote_scan = 0x16
    emote_sleep = 0x17
    emote_surprised = 0x18
    idle_1 = 0x19
    idle_2 = 0x1a
    idle_3 = 0x1b
    patrol_alarm = 0x1c
    patrol_hit = 0x1d
    patrol_patrolling = 0x1e
    wwm_angry = 0x1f
    wwm_anxious = 0x20
    wwm_bow = 0x21
    wwm_concern = 0x22
    wwm_curious = 0x23
    wwm_double_take = 0x24
    wwm_excited = 0x25
    wwm_fiery = 0x26
    wwm_frustrated = 0x27
    wwm_happy = 0x28
    wwm_jittery = 0x29
    wwm_laugh = 0x2a
    wwm_long_shake = 0x2b
    wwm_no = 0x2c
    wwm_ominous = 0x2d
    wwm_relieved = 0x2e
    wwm_sad = 0x2f
    wwm_scared = 0x30
    wwm_shake = 0x31
    wwm_surprised = 0x32
    wwm_taunting = 0x33
    wwm_whisper = 0x34
    wwm_yelling = 0x35
    wwm_yoohoo = 0x36
    motor = 0x37


class R2Q5Animation(Enum):
    charger_1 = 0x00
    charger_3 = 0x01
    charger_2 = 0x02
    charger_4 = 0x03
    charger_5 = 0x04
    charger_6 = 0x05
    charger_7 = 0x06
    emote_alarm = 0x07
    emote_angry = 0x08
    emote_attention = 0x09
    emote_frustrated = 0x0a
    emote_drive = 0x0b
    emote_excited = 0x0c
    emote_search = 0x0d
    emote_short_circuit = 0x0e
    emote_laugh = 0x0f
    emote_no = 0x10
    emote_retreat = 0x11
    emote_fiery = 0x12
    emote_understood = 0x13
    emote_yes = 0x15
    emote_scan = 0x16
    emote_surprised = 0x18
    idle_1 = 0x19
    idle_2 = 0x1a
    idle_3 = 0x1b
    wwm_angry = 0x1f
    wwm_anxious = 0x20
    wwm_bow = 0x21
    wwm_concern = 0x22
    wwm_curious = 0x23
    wwm_double_take = 0x24
    wwm_excited = 0x25
    wwm_fiery = 0x26
    wmm_frustrated = 0x27
    wwm_happy = 0x28
    wwm_jittery = 0x29
    wwm_laugh = 0x2a
    wwm_long_shake = 0x2b
    wwm_no = 0x2c
    wwm_ominous = 0x2d
    wwm_relieved = 0x2e
    wwm_sad = 0x2f
    wwm_scared = 0x30
    wwm_shake = 0x31
    wwm_surprised = 0x32
    wwm_taunting = 0x33
    wwm_whisper = 0x34
    wwm_yelling = 0x35
    wwm_yoohoo = 0x36


class BB9EAnimation(Enum):
    EMOTE_ALARM = 0x00
    EMOTE_NO = 0x01
    EMOTE_SCAN_SWEEP = 0x02
    EMOTE_SCARED = 0x03
    EMOTE_YES = 0x04
    EMOTE_AFFIRMATIVE = 0x05
    EMOTE_AGITATED = 0x06
    EMOTE_ANGRY = 0x07
    EMOTE_CONTENT = 0x08
    EMOTE_EXCITED = 0x09
    EMOTE_FIERY = 0x0a
    EMOTE_GREETINGS = 0x0b
    EMOTE_NERVOUS = 0x0c
    EMOTE_SLEEP = 0x0e
    EMOTE_SURPRISED = 0x0f
    EMOTE_UNDERSTOOD = 0x10
    HIT = 0x11
    WWM_ANGRY = 0x12
    WWM_ANXIOUS = 0x13
    WWM_BOW = 0x14
    WWM_CURIOUS = 0x16
    WWM_DOUBLE_TAKE = 0x17
    WWM_EXCITED = 0x18
    WWM_FIERY = 0x19
    WWM_HAPPY = 0x1a
    WWM_JITTERY = 0x1b
    WWM_LAUGH = 0x1c
    WWM_LONG_SHAKE = 0x1d
    WWM_NO = 0x1e
    WWM_OMINOUS = 0x1f
    WWM_RELIEVED = 0x20
    WWM_SAD = 0x21
    WWM_SCARED = 0x22
    WWM_SHAKE = 0x23
    WWM_SURPRISED = 0x24
    WWM_TAUNTING = 0x25
    WWM_WHISPER = 0x26
    WWM_YELLING = 0x27
    WWM_YOOHOO = 0x28
    WWM_FRUSTRATED = 0x29
    IDLE_1 = 0x2a
    IDLE_2 = 0x2b
    IDLE_3 = 0x2c
    EYE_1 = 0x2d
    EYE_2 = 0x2e
    EYE_3 = 0x2f
    EYE_4 = 0x30


class R2LegAction(Enum):
    stop = 0x00
    three_legs = 0x01
    two_legs = 0x02
    waddle = 0x03

class AnimatronicsLMQCommand(Enum):
    animate_eyes = 0x01
    animate_suspension = 0x03
    animate_mouth = 0x04
    execute_animation_profile = 0x05
    upload_animation_bundle = 0x06
    capacitive_touch_indication = 0x07
    upload_animation_bundle_result = 0x09
    enable_idle_animation = 0x0a
    enable_user_drive = 0x0b
    do_shoulder_action = 0x0d
    set_head_position = 0x0f
    execute_animation_bundle_complete = 0x11
    animate_steering = 0x12
    animation_upload_process_complete = 0x13
    get_head_position = 0x14
    set_shoulder_cam_position = 0x15
    get_shoulder_cam_position = 0x16
    set_shoulder_action = 0x25
    shoulder_action_complete = 0x26
    enable_shoulder_action_complete = 0x2a
    stop_animation = 0x2b

# ALL LMQUXXX have been found like that, sentences written are the french one
# will switch my device to english to provide the proper sentence, but if someone could
# do it will help.
class LMQAnimation(Enum):
    Oh_yeah_Lightning_is_ready_LMQ007 = 0x24
    Bien_joue_LMQU006 = 0x26
    Cool_LMQ012 = 0x28
    Thank_you_LMQ016 = 0x2b
    This_is_amazing_LMQ024 = 0x32
    Whoa_LMQ177 = 0x33
    Hey_c_mon_let_s_go_LMQ028 = 0x36
    Have_I_mentioned_that_I_love_your_garage_LMQ031 = 0x39
    Look_out_folks_Lightning_McQueen_is_back_LMQ039 = 0x3f
    I_am_so_ready_for_the_Piston_Cup_LMQ044 = 0x41
    Look_left_look_right_Yep_nothing_but_open_road_LMQ062 = 0x4e
    My_buddy_Filmore_makes_the_best_organic_fuel_in_the_world_loads_of_power_and_a_clean_burn_LMQ067 = 0x52
    The_road_is_calling_my_name_Hear_it_Lightning_oh_Lightning_LMQ069 = 0x54
    I_am_speed_LMQ432 = 0x5e  # je suis rapide
    I_ve_been_all_over_the_world_and_I_can_honestly_say_that_Flo_s_V_8_Cafe_serves_the_best_quart_of_oil_anywhere_LMQ074 = 0x58
    Crank_the_acceleration_LMQ083 = 0x59
    Focus_focus_LMQ089 = 0x5f
    The_crowd_loves_us_LMQ100 = 0x64
    Game_on_LMQ105 = 0x66
    Next_up_Lightning_McQueen_LMQ108 = 0x67
    Ka_Chow_LMQ113 = 0x69
    Yes_woo_LMQ116 = 0x6c
    Pay_attention_this_is_how_it_s_done_LMQ143 = 0x7b
    Je_suis_Flash_Mcqueen_LMQU008 = 0x7e
    Wow_LMQ025 = 0x82
    Ho_oui_Flash_Mcqueen_est_dans_la_place_LMQU003 = 0x8d
    Hey_LMQ237 = 0x9f
    Here_I_go_LMQ238 = 0xa0
    Woo_hoo_LMQ241 = 0xa2
    Ah_serieux_on_devrait_essayer_autre_chose_LMQU005 = 0xcd
    Ahhh_LMQ379 = 0xdc
    Tu_maitrises_mon_ami_LMQU010 = 0xce
    Gasp_LMQ403 = 0xf2
    Je_n_ai_jamais_ete_aussi_bien_prepare_LMQU009 = 0xf5
    Ah_Quels_souvenirs_LMQU007 = 0xf6
    J_adore_quand_elles_m_appellent_chouchou_LMQ008 = 0xfe
    Watch_this_I_call_it_The_Wave_LMQ431 = 0x107
    Groovin_LMQ457 = 0x111
    Watch_me_now_LMQ458 = 0x112
    Entendu_allons_y_je_veux_bien_passer_devant_la_camera_LMQU001 = 0x126
    Etre_un_champion_repose_sur_le_travail_d_equipe_ce_qui_implique_d_avoir_un_stand_de_ravitaillement_d_execption_LMQU002 = 0x128
    Je_suis_super_rapide_LMQU004 = 0x12c
    I_m_a_precision_instrument_of_speed_and_aerodynamics_LMQ433 = 0x12d
    Ahh_c_mon_LMQ125 = 0x12f
    Engine_Rev_LMQ_Start_2 = 0x130
    Engine_Start_LMQ_Start_3 = 0x131
    Engine_Idle_LMQ_Idle_Loop = 0x132

    # sentence in the app that not captured yet.
    # Woooo_Woo_What_a_blast_LMQ041 =
    # When_it_comes_to_medicated_bumper_ointment_Rust-eze_is_the_only_brand_I_trust_LMQ073 =
    # Dahh_LMQ394 =


class Animatronics(DeviceApiABC):
    device_id = DeviceId.animatronics

    def __init__(self, ble_adapter):
        super().__init__(ble_adapter)
        self._wait_for_play_animation = False
        self._animation: Optional[List[int]] = None

    def play_animation(self, animation_id: int, target_id=0x12):
        self.request(
            AnimatronicsCommand.play_animation,
            data=[*animation_id.to_bytes(2, "big")],
            target_id=target_id,
        )

    def play_animation_and_wait(self, animation_id: int, target_id=0x12, timeout: float = 10):
        self._animation = [*animation_id.to_bytes(2, "big")]
        self._wait_for_play_animation = True

        def callback_wrapper(response: Packet):
            if response.data == self._animation:
                self._wait_for_play_animation = False

        self.notify(AnimatronicsCommand.play_animation_complete_notify, callback_wrapper, timeout=timeout)
        self.request(
            AnimatronicsCommand.play_animation,
            data=self._animation,
            target_id=target_id,
        )

        sleep_time = 0.1
        while self._wait_for_play_animation:
            timeout -= sleep_time
            if timeout <= 0:
                self._wait_for_play_animation = False
            time.sleep(sleep_time)

        self.cancel_notify()

    def perform_leg_action(self, leg_action: R2LegAction):
        self.request(
            AnimatronicsCommand.perform_leg_action,
            data=[*leg_action.value.to_bytes(2, "big")],
            target_id=0x12,
        )

    def set_head_position(self, position: float):
        self.request(
            AnimatronicsCommand.set_head_position,
            data=[*struct.pack("f", position)],
            target_id=0x12,
        )

    def get_head_position(self) -> float:
        response = self.request(
            AnimatronicsCommand.get_head_position,
            target_id=0x12,
        )
        return float_from_bytes(response.data)

    def set_leg_position(self, position: float):
        self.request(
            AnimatronicsCommand.set_leg_position,
            data=[*struct.pack("f", position)],
            target_id=0x12,
        )

    def get_leg_position(self) -> float:
        response = self.request(
            AnimatronicsCommand.get_leg_position,
            target_id=0x12,
        )
        return float_from_bytes(response.data)

    def get_leg_action(self) -> R2LegAction:
        response = self.request(
            AnimatronicsCommand.get_leg_action,
            target_id=0x12,
        )
        return R2LegAction(int.from_bytes(response.data[:2], "big"))

    def stop_animation(self):
        self.request(
            AnimatronicsCommand.stop_animation,
            target_id=0x12,
        )

    def get_trophy_mode_enabled(self) -> bool:
        response = self.request(
            AnimatronicsCommand.get_trophy_mode_enabled,
            target_id=0x12,
        )
        return bool(int.from_bytes(response.data, "big"))



from setuptools import setup, find_packages

__pckg__ = "pysphero"
__dpckg__ = __pckg__.replace("-", "_")
__version__ = "0.0.12"


def long_description():
    with open("README.md") as readme:
        return readme.read()


setup(
    name=__pckg__,
    version=__version__,
    description="Unofficial library for sphero v2.",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author="Andrey Lemets",
    author_email="a.a.lemets@gmail.com",
    url="https://github.com/EnotYoyo/pysphero",
    packages=find_packages(),
    include_package_data=True,
    license="MIT License",
    install_requires=[
        "bluepy==1.3.0",
    ],
    extras_require={
        "tests": [
            "pytest==4.2.0",
            "pytest-cov==2.6.1",
        ],
        "gatt": [
            "gatt==0.2.7",
        ]
    },
    keywords=["sphero", "sphero-ble", "bolt"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



from enum import Enum


class SecondaryMcuFirmwareUpdateCommand(Enum):
    begin_reflash = 0x00
    here_is_page = 0x01
    jump_to_main_app = 0x02
    is_page_blank = 0x03
    erase_user_config = 0x04
    jump_to_bootloader = 0x05
    prepare_for_firmware_update = 0x06
    ready_for_firmware_update = 0x07
    complete_firmware_update = 0x08
    firmware_update_complete = 0x09
    application_identification = 0x0a
    request_application_id = 0x0b
    set_pending_update_flags = 0x0c
    get_pending_update_flags = 0x0d



import logging
import time
from typing import List, Optional

from pysphero.constants import Api2Error
from pysphero.exceptions import PySpheroTimeoutError, PySpheroRuntimeError, PySpheroApiError
from pysphero.packet import Packet, Flag

logger = logging.getLogger(__name__)


class PacketCollector:
    def __init__(self, check_response_delta: float = 0.1):
        self._data = []
        self._packets = {}
        self._check_response_delta = check_response_delta or 0.01  # protect of 0 sleep_time

    def append_raw_data(self, data: List[int]):
        for b in data:
            logger.debug(f"Received {b:#04x}")
            self._data.append(b)

            # packet always ending with end byte
            if b == Packet.end:
                if len(self._data) < 6:
                    raise PySpheroRuntimeError(f"Very small packet {[hex(x) for x in self.data]}")
                self._build_packet()

    def _build_packet(self):
        """
        Create and save packet from raw bytes
        """
        logger.debug(f"Starting of packet build")

        packet = Packet.from_response(self._data)
        self._packets[packet.id] = packet
        self._data = []

    def get_response(self, packet: Packet, raise_api_error: bool = True, timeout: float = 10) -> Optional[Packet]:
        if (packet.flags & (Flag.requests_response.value | Flag.requests_only_error_response.value)) == 0:
            return

        while True:
            response = self._packets.pop(packet.id, None)
            if response:
                if raise_api_error and response.api_error is not Api2Error.success:
                    raise PySpheroApiError(response.api_error)

                return response

            timeout -= self._check_response_delta
            if timeout <= 0:
                raise PySpheroTimeoutError(f"Timeout error for response of {packet}")

            time.sleep(self._check_response_delta)
